#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import ArgumentParser, Namespace
from typing import List

from idb.cli.commands.base import TargetCommand
from idb.client.client import IdbClient
from idb.common.signal import signal_handler_event


class RecordVideoCommand(TargetCommand):
    @property
    def description(self) -> str:
        return "Record the target's screen to a mp4 video file"

    @property
    def name(self) -> str:
        return "video"

    @property
    def aliases(self) -> List[str]:
        return ["record-video"]

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("output_file", help="mp4 file to output the video to")
        super().add_parser_arguments(parser)

    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
        await client.record_video(
            stop=signal_handler_event("video"), output_file=args.output_file
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import AsyncIterable, AsyncIterator, Dict, Iterable, List, Optional, Tuple

from idb.grpc.types import CompanionClient
from idb.grpc.stream import drain_to_stream
from idb.common.types import HIDButtonType
from idb.grpc.idb_pb2 import HIDEvent, Point


HIDButton = HIDEvent.HIDButton
HIDDelay = HIDEvent.HIDDelay
HIDKey = HIDEvent.HIDKey
HIDPress = HIDEvent.HIDPress
HIDPressAction = HIDEvent.HIDPressAction
HIDSwipe = HIDEvent.HIDSwipe
HIDTouch = HIDEvent.HIDTouch


def tap_to_events(x: int, y: int, duration: Optional[float] = None) -> List[HIDEvent]:
    return _press_with_duration(
        HIDPressAction(touch=HIDTouch(point=Point(x=x, y=y))), duration=duration
    )


def button_press_to_events(
    button: HIDButtonType, duration: Optional[float] = None
) -> List[HIDEvent]:
    # Need to convert between the py enum that starts at 1 and the grpc enum
    # that starts at 0
    return _press_with_duration(
        HIDPressAction(button=HIDButton(button=_translate_button_type(button))),
        duration=duration,
    )


def _translate_button_type(button: HIDButtonType) -> HIDEvent.HIDButtonType:
    if button == HIDButtonType.APPLE_PAY:
        return HIDEvent.APPLE_PAY
    elif button == HIDButtonType.HOME:
        return HIDEvent.HOME
    elif button == HIDButtonType.LOCK:
        return HIDEvent.LOCK
    elif button == HIDButtonType.SIDE_BUTTON:
        return HIDEvent.SIDE_BUTTON
    elif button == HIDButtonType.SIRI:
        return HIDEvent.SIRI
    raise Exception(f"Unexpected button type {button}")


def key_press_to_events(
    keycode: int, duration: Optional[float] = None
) -> List[HIDEvent]:
    return _press_with_duration(
        HIDPressAction(key=HIDKey(keycode=keycode)), duration=duration
    )


def _press_with_duration(
    action: HIDPressAction, duration: Optional[float] = None
) -> List[HIDEvent]:
    events = []
    events.append(HIDEvent(press=HIDPress(action=action, direction=HIDEvent.DOWN)))
    if duration:
        events.append(HIDEvent(delay=HIDDelay(duration=duration)))
    events.append(HIDEvent(press=HIDPress(action=action, direction=HIDEvent.UP)))
    return events


def swipe_to_events(
    p_start: Tuple[float, float],
    p_end: Tuple[float, float],
    delta: Optional[float] = None,
) -> List[HIDEvent]:
    start = Point(x=p_start[0], y=p_start[1])
    end = Point(x=p_end[0], y=p_end[1])
    return [HIDEvent(swipe=HIDSwipe(start=start, end=end, delta=delta))]


def _key_down_event(keycode: int) -> HIDEvent:
    return HIDEvent(
        press=HIDPress(
            action=HIDPressAction(key=HIDKey(keycode=keycode)), direction=HIDEvent.DOWN
        )
    )


def _key_up_event(keycode: int) -> HIDEvent:
    return HIDEvent(
        press=HIDPress(
            action=HIDPressAction(key=HIDKey(keycode=keycode)), direction=HIDEvent.UP
        )
    )


def key_press_shifted_to_events(keycode: int) -> List[HIDEvent]:
    return [
        _key_down_event(225),
        _key_down_event(keycode),
        _key_up_event(keycode),
        _key_up_event(225),
    ]


KEY_MAP: Dict[str, List[HIDEvent]] = {
    "a": key_press_to_events(4),
    "b": key_press_to_events(5),
    "c": key_press_to_events(6),
    "d": key_press_to_events(7),
    "e": key_press_to_events(8),
    "f": key_press_to_events(9),
    "g": key_press_to_events(10),
    "h": key_press_to_events(11),
    "i": key_press_to_events(12),
    "j": key_press_to_events(13),
    "k": key_press_to_events(14),
    "l": key_press_to_events(15),
    "m": key_press_to_events(16),
    "n": key_press_to_events(17),
    "o": key_press_to_events(18),
    "p": key_press_to_events(19),
    "q": key_press_to_events(20),
    "r": key_press_to_events(21),
    "s": key_press_to_events(22),
    "t": key_press_to_events(23),
    "u": key_press_to_events(24),
    "v": key_press_to_events(25),
    "w": key_press_to_events(26),
    "x": key_press_to_events(27),
    "y": key_press_to_events(28),
    "z": key_press_to_events(29),
    "A": key_press_shifted_to_events(4),
    "B": key_press_shifted_to_events(5),
    "C": key_press_shifted_to_events(6),
    "D": key_press_shifted_to_events(7),
    "E": key_press_shifted_to_events(8),
    "F": key_press_shifted_to_events(9),
    "G": key_press_shifted_to_events(10),
    "H": key_press_shifted_to_events(11),
    "I": key_press_shifted_to_events(12),
    "J": key_press_shifted_to_events(13),
    "K": key_press_shifted_to_events(14),
    "L": key_press_shifted_to_events(15),
    "M": key_press_shifted_to_events(16),
    "N": key_press_shifted_to_events(17),
    "O": key_press_shifted_to_events(18),
    "P": key_press_shifted_to_events(19),
    "Q": key_press_shifted_to_events(20),
    "R": key_press_shifted_to_events(21),
    "S": key_press_shifted_to_events(22),
    "T": key_press_shifted_to_events(23),
    "U": key_press_shifted_to_events(24),
    "V": key_press_shifted_to_events(25),
    "W": key_press_shifted_to_events(26),
    "X": key_press_shifted_to_events(27),
    "Y": key_press_shifted_to_events(28),
    "Z": key_press_shifted_to_events(29),
    "1": key_press_to_events(30),
    "2": key_press_to_events(31),
    "3": key_press_to_events(32),
    "4": key_press_to_events(33),
    "5": key_press_to_events(34),
    "6": key_press_to_events(35),
    "7": key_press_to_events(36),
    "8": key_press_to_events(37),
    "9": key_press_to_events(38),
    "0": key_press_to_events(39),
    ";": key_press_to_events(51),
    "=": key_press_to_events(46),
    ",": key_press_to_events(54),
    "-": key_press_to_events(45),
    ".": key_press_to_events(55),
    "/": key_press_to_events(56),
    "`": key_press_to_events(53),
    "[": key_press_to_events(47),
    "\\": key_press_to_events(49),
    "]": key_press_to_events(48),
    "'": key_press_to_events(52),
    " ": key_press_to_events(44),
    "!": key_press_shifted_to_events(30),
    "@": key_press_shifted_to_events(31),
    "#": key_press_shifted_to_events(32),
    "$": key_press_shifted_to_events(33),
    "%": key_press_shifted_to_events(34),
    "^": key_press_shifted_to_events(35),
    "&": key_press_shifted_to_events(36),
    "*": key_press_shifted_to_events(37),
    "(": key_press_shifted_to_events(38),
    ")": key_press_shifted_to_events(39),
    "_": key_press_shifted_to_events(45),
    "+": key_press_shifted_to_events(46),
    "{": key_press_shifted_to_events(47),
    "}": key_press_shifted_to_events(48),
    ":": key_press_shifted_to_events(51),
    '"': key_press_shifted_to_events(52),
    "|": key_press_shifted_to_events(49),
    "<": key_press_shifted_to_events(54),
    ">": key_press_shifted_to_events(55),
    "?": key_press_shifted_to_events(56),
    "~": key_press_shifted_to_events(53),
}


def text_to_events(text: str) -> List[HIDEvent]:
    events = []
    for character in text:
        if character in KEY_MAP:
            events.extend(KEY_MAP[character])
        else:
            raise Exception(f"No keycode found for {character}")
    return events


async def _iterator_to_async_iterator(
    events: Iterable[HIDEvent]
) -> AsyncIterator[HIDEvent]:
    for event in events:
        yield event


async def send_events(client: CompanionClient, events: Iterable[HIDEvent]) -> None:
    await hid(client, _iterator_to_async_iterator(events))


async def tap(
    client: CompanionClient, x: int, y: int, duration: Optional[float] = None
) -> None:
    await send_events(client, tap_to_events(x, y, duration))


async def button(
    client: CompanionClient,
    button_type: HIDButtonType,
    duration: Optional[float] = None,
) -> None:
    await send_events(client, button_press_to_events(button_type, duration))


async def key(
    client: CompanionClient, keycode: int, duration: Optional[float] = None
) -> None:
    await send_events(client, key_press_to_events(keycode, duration))


async def text(client: CompanionClient, text: str) -> None:
    await send_events(client, text_to_events(text))


async def swipe(
    client: CompanionClient,
    p_start: Tuple[int, int],
    p_end: Tuple[int, int],
    delta: Optional[int] = None,
) -> None:
    await send_events(client, swipe_to_events(p_start, p_end, delta))


async def key_sequence(client: CompanionClient, key_sequence: List[int]) -> None:
    events: List[HIDEvent] = []
    for key in key_sequence:
        events.extend(key_press_to_events(key))
    await send_events(client, events)


async def hid(client: CompanionClient, event_iterator: AsyncIterable[HIDEvent]) -> None:
    async with client.stub.hid.open() as stream:
        await drain_to_stream(
            stream=stream, generator=event_iterator, logger=client.logger
        )
        await stream.recv_message()


CLIENT_PROPERTIES = [tap, button, key, key_sequence, text, swipe, hid]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
import json
import logging
import os
from asyncio import StreamReader
from asyncio.subprocess import Process
from typing import List, Tuple

from idb.common.constants import IDB_LOGS_PATH


class CompanionSpawnerException(Exception):
    pass


class CompanionSpawner:
    def __init__(self, companion_path: str) -> None:
        self.companion_path = companion_path
        self.companion_processes: List[Process] = []

    async def _read_stream(self, stream: StreamReader) -> int:
        port = 0
        while True:
            line = await stream.readline()
            if line:
                update = json.loads(line.decode())
                if update:
                    logging.debug(f"got update from companion {update}")
                    port = update["grpc_port"]
                    break
            else:
                break
        return port

    def _log_file_path(self, target_udid: str) -> str:
        os.makedirs(name=IDB_LOGS_PATH, exist_ok=True)
        return IDB_LOGS_PATH + "/" + target_udid

    async def spawn_companion(self, target_udid: str) -> int:
        if not self.companion_path:
            raise CompanionSpawnerException(
                f"couldn't instantiate a companion for {target_udid} because\
                 the companion_path is not available"
            )
        cmd: List[str] = [
            self.companion_path,
            "--udid",
            target_udid,
            "--grpc-port",
            "0",
        ]

        with open(self._log_file_path(target_udid), "a") as log_file:
            process = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stdin=asyncio.subprocess.PIPE,
                stderr=log_file,
            )
            self.companion_processes.append(process)
            logging.debug(f"started companion at process id {process.pid}")
            if process.stdout:
                return await self._read_stream(process.stdout)
            raise CompanionSpawnerException("process has no stdout")

    def close(self) -> None:
        logging.info("Stopping companion spawner")
        for process in self.companion_processes:
            logging.info("Stopping companion")
            process.terminate()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.grpc.types import CompanionClient
from idb.grpc.idb_pb2 import MvRequest, MvResponse


async def client(
    client: CompanionClient, bundle_id: str, src_paths: List[str], dest_path: str
) -> None:
    await client.stub.mv(
        MvRequest(bundle_id=bundle_id, src_paths=src_paths, dst_path=dest_path)
    )


async def daemon(client: CompanionClient, request: MvRequest) -> MvResponse:
    return await client.stub.mv(request)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from argparse import ArgumentParser, REMAINDER, Namespace
from typing import List, Optional, Set

from idb.cli.commands.base import Command, CompositeCommand, TargetCommand
from idb.client.client import IdbClient
from idb.common.format import human_format_test_info, json_format_test_info
from idb.common.misc import get_env_with_idb_prefix


class CommonRunXcTestCommand(TargetCommand):
    @property
    def description(self) -> str:
        return (
            f"Run an installed {self.name} test. Will pass through"
            " any environment\nvariables prefixed with IDB_"
        )

    def add_parser_positional_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "test_bundle_id", help="Bundle id of the test to launch", type=str
        )

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        self.add_parser_positional_arguments(parser)
        parser.add_argument(
            "--result-bundle-path",
            default=None,
            type=str,
            help="Path to save the result bundle",
        )
        parser.add_argument(
            "--timeout", help="Seconds before timeout occurs", default=3600, type=int
        )
        parser.add_argument(
            "test_arguments",
            help="Arguments to start the test with",
            default=[],
            nargs=REMAINDER,
        )
        super().add_parser_arguments(parser)

    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
        await super().run_with_client(args, client)
        tests_to_run = self.get_tests_to_run(args)
        tests_to_skip = self.get_tests_to_skip(args)
        app_bundle_id = args.app_bundle_id if hasattr(args, "app_bundle_id") else None
        test_host_app_bundle_id = (
            args.test_host_app_bundle_id
            if hasattr(args, "test_host_app_bundle_id")
            else None
        )
        is_ui = args.run == "ui"
        is_logic = args.run == "logic"

        formatter = json_format_test_info if args.json else human_format_test_info
        async for test_result in client.run_xctest(
            test_bundle_id=args.test_bundle_id,
            app_bundle_id=app_bundle_id,
            test_host_app_bundle_id=test_host_app_bundle_id,
            is_ui_test=is_ui,
            is_logic_test=is_logic,
            tests_to_run=tests_to_run,
            tests_to_skip=tests_to_skip,
            timeout=args.timeout,
            env=get_env_with_idb_prefix(),
            args=args.test_arguments,
            result_bundle_path=args.result_bundle_path,
        ):
            print(formatter(test_result))

    def get_tests_to_run(self, args: Namespace) -> Optional[Set[str]]:
        return None

    def get_tests_to_skip(self, args: Namespace) -> Optional[Set[str]]:
        return None


class RunXcTestAppCommand(CommonRunXcTestCommand):
    @property
    def name(self) -> str:
        return "app"

    def add_parser_positional_arguments(self, parser: ArgumentParser) -> None:
        super().add_parser_positional_arguments(parser)
        parser.add_argument(
            "app_bundle_id", help="Bundle id of the app to test", type=str
        )

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        super().add_parser_arguments(parser)
        parser.add_argument(
            "--tests-to-run",
            nargs="*",
            help="Run only these tests, \
            if not specified all tests are run. \
            Format: className/methodName",
        )
        parser.add_argument(
            "--tests-to-skip",
            nargs="*",
            help="Skip these tests, \
            has precedence over --tests-to-run. \
            Format: className/methodName",
        )

    def get_tests_to_run(self, args: Namespace) -> Optional[Set[str]]:
        return set(args.tests_to_run) if args.tests_to_run else None

    def get_tests_to_skip(self, args: Namespace) -> Optional[Set[str]]:
        return set(args.tests_to_skip) if args.tests_to_skip else None


class RunXcTestUICommand(RunXcTestAppCommand):
    @property
    def name(self) -> str:
        return "ui"

    def add_parser_positional_arguments(self, parser: ArgumentParser) -> None:
        super().add_parser_positional_arguments(parser)
        parser.add_argument(
            "test_host_app_bundle_id",
            help="Bundle id of the app that hosts ui test",
            type=str,
        )


class RunXcTestLogicCommand(CommonRunXcTestCommand):
    @property
    def name(self) -> str:
        return "logic"

    def add_parser_arguments(self, parser: ArgumentParser) -> None:
        super().add_parser_arguments(parser)
        parser.add_argument(
            "--test-to-run",
            nargs=1,
            help="Run only this test, \
            if not specified all tests are run. \
            Format: className",
        )

    def get_tests_to_run(self, args: Namespace) -> Optional[Set[str]]:
        return set(args.test_to_run) if args.test_to_run else None


class RunXctestCommand(CompositeCommand):
    def __init__(self) -> None:
        self._subcommands: List[Command] = [
            RunXcTestAppCommand(),
            RunXcTestUICommand(),
            RunXcTestLogicCommand(),
        ]

    @property
    def subcommands(self) -> List[Command]:
        return self._subcommands

    @property
    def description(self) -> str:
        return (
            "Run an installed xctest. Will pass through any environment\n"
            "variables prefixed with IDB_"
        )

    @property
    def name(self) -> str:
        return "run"

    async def run(self, args: Namespace, context: Optional[Namespace] = None) -> None:
        await super().run(args, context)  # pyre-ignore

    async def run_with_client(self, args: Namespace, client: IdbClient) -> None:
        await self.run(args)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from idb.grpc.types import CompanionClient
from idb.common.tar import create_tar
from idb.grpc.idb_pb2 import ContactsUpdateRequest, Payload


async def client(client: CompanionClient, contacts_path: str) -> None:
    data = await create_tar([contacts_path])
    await client.stub.contacts_update(ContactsUpdateRequest(payload=Payload(data=data)))



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


from idb.grpc.types import CompanionClient
from idb.grpc.idb_pb2 import ScreenshotRequest


async def client(client: CompanionClient) -> bytes:
    response = await client.stub.screenshot(ScreenshotRequest())
    return response.image_data



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


import asyncio
from typing import AsyncIterator, List, Optional

from idb.grpc.idb_pb2 import LogRequest
from idb.grpc.stream import cancel_wrapper
from idb.grpc.types import CompanionClient


async def tail_logs(
    client: CompanionClient, stop: asyncio.Event, arguments: Optional[List[str]] = None
) -> AsyncIterator[str]:
    async with client.stub.log.open() as stream:
        await stream.send_message(LogRequest(arguments=arguments), end=True)
        async for message in cancel_wrapper(stream=stream, stop=stop):
            yield message.output.decode()


CLIENT_PROPERTIES = [tail_logs]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.


from typing import Set, Dict  # noqa F401

from idb.grpc.types import CompanionClient
from idb.grpc.idb_pb2 import ApproveRequest


MAP = {  # type: Dict[str, ApproveRequest.Permission]
    "photos": ApproveRequest.PHOTOS,
    "camera": ApproveRequest.CAMERA,
    "contacts": ApproveRequest.CONTACTS,
}


async def client(
    client: CompanionClient, bundle_id: str, permissions: Set[str]
) -> None:
    print(f"Sending {[MAP[permission] for permission in permissions]}")
    await client.stub.approve(
        ApproveRequest(
            bundle_id=bundle_id,
            permissions=[MAP[permission] for permission in permissions],
        )
    )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.manager.companion import CompanionClient
from idb.grpc.ipc_loader import DaemonContext
from idb.common.types import TargetDescription
from idb.grpc.idb_pb2 import ListTargetsRequest, ListTargetsResponse
from idb.ipc.mapping.target import target_to_grpc, target_to_py


async def client(client: CompanionClient) -> List[TargetDescription]:
    response = await client.stub.list_targets(ListTargetsRequest())
    return [target_to_py(target) for target in response.targets]


async def daemon(
    context: DaemonContext, request: ListTargetsRequest
) -> ListTargetsResponse:
    targets = context.companion_manager.targets
    return ListTargetsResponse(targets=[target_to_grpc(target) for target in targets])



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.grpc.types import CompanionClient
from idb.common.types import FileEntryInfo
from idb.grpc.idb_pb2 import LsRequest, LsResponse


async def client(
    client: CompanionClient, bundle_id: str, path: str
) -> List[FileEntryInfo]:
    response = await client.stub.ls(LsRequest(bundle_id=bundle_id, path=path))
    return [FileEntryInfo(path=file.path) for file in response.files]


async def daemon(client: CompanionClient, request: LsRequest) -> LsResponse:
    return await client.stub.ls(request)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from typing import List

from idb.manager.companion import CompanionClient
from idb.grpc.idb_pb2 import XctestListTestsRequest


async def list_test_bundle(client: CompanionClient, test_bundle_id: str) -> List[str]:
    response = await client.stub.xctest_list_tests(
        XctestListTestsRequest(bundle_name=test_bundle_id)
    )
    return [name for name in response.names]


CLIENT_PROPERTIES = [list_test_bundle]  # pyre-ignore



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import inspect
import logging
from typing import Any, Callable, Optional, Set

from idb.client.daemon_pid_saver import kill_saved_pids
from idb.common.types import IdbClientBase


BASE_MEMBERS: Set[str] = {
    name
    for (name, value) in inspect.getmembers(IdbClientBase)
    if not name.startswith("__")
}


class IdbClient(IdbClientBase):
    def __init__(
        self,
        resolve: Callable[[str], IdbClientBase],
        logger: Optional[logging.Logger] = None,
    ) -> None:
        self.logger: logging.Logger = logger or logging.getLogger("idb_client")
        self._resolve = resolve

    def __getattribute__(self, key: str) -> Any:  # pyre-ignore
        if key not in BASE_MEMBERS:
            return super().key
        logger = super().__getattribute__("logger")
        logger.debug(f"Resolving client for {key}")
        client = super().__getattribute__("_resolve")(key)
        logger.debug(f"Resolved client for {key} to {client}")
        return getattr(client, key)

    @classmethod
    async def kill(cls) -> None:
        await kill_saved_pids()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from idb.grpc.types import CompanionClient
from idb.grpc.ipc_loader import DaemonContext
from idb.common.types import ConnectionDestination
from idb.grpc.idb_pb2 import DisconnectRequest, DisconnectResponse
from idb.ipc.mapping.destination import destination_to_grpc, destination_to_py
from idb.utils.typing import none_throws


async def client(
    client: CompanionClient,
    destination: ConnectionDestination,
    disconnect_from_daemon: bool = False,
) -> None:
    await client.stub.disconnect(
        DisconnectRequest(
            destination=destination_to_grpc(destination),
            is_daemon=disconnect_from_daemon,
        )
    )


async def daemon(
    context: DaemonContext, request: DisconnectRequest
) -> DisconnectResponse:
    destination = destination_to_py(none_throws(request.destination))
    context.companion_manager.remove_companion(destination)
    return DisconnectResponse()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import json
from unittest import mock
from typing import List

from idb.common.types import TargetDescription
from idb.daemon.companion_tailer import CompanionTailer
from idb.manager.companion import CompanionManager
from idb.utils.testing import AsyncMock, TestCase, ignoreTaskLeaks


@ignoreTaskLeaks
class CompanionTailerTest(TestCase):
    async def test_spawn_companion(self) -> None:
        tailer = CompanionTailer("idb_path", mock.Mock())
        tailer._log_file_path = mock.Mock()
        with mock.patch(
            "idb.daemon.companion_tailer.asyncio.create_subprocess_exec",
            new=AsyncMock(),
        ) as exec_mock, mock.patch("idb.daemon.companion_tailer.open"):
            process_mock = mock.Mock()
            process_mock.stdout.readline = AsyncMock(return_value=None)
            exec_mock.return_value = process_mock
            await tailer.start()
            exec_mock.assert_called_once_with(
                "idb_path", "--notify", "1", stdout=mock.ANY, stderr=mock.ANY
            )
            self.assertEqual(tailer.process, process_mock)

    async def test_close(self) -> None:
        tailer = CompanionTailer("idb_path", mock.Mock())
        process_mock = mock.Mock()
        tailer.process = process_mock
        tailer._reading_forever_fut = mock.Mock()
        tailer.close()
        process_mock.terminate.assert_called_once()
        tailer._reading_forever_fut.cancel.assert_called_once()

    async def test_read_stream(self) -> None:
        class StreamMock:
            i = 0
            lines: List[bytes] = [
                json.dumps(
                    {
                        "udid": "udid",
                        "state": "state",
                        "type": "type",
                        "name": "name",
                        "os_version": "os_version",
                        "architecture": "architecture",
                    }
                ).encode("utf-8"),
                json.dumps({"initial_state_ended": True}).encode("utf-8"),
            ]

            async def readline(self) -> bytes:
                result = self.lines[self.i]
                self.i += 1
                return result

        manager = CompanionManager(None, mock.MagicMock())
        tailer = CompanionTailer("idb_path", manager)
        await tailer._read_stream(StreamMock())  # pyre-ignore
        self.assertEqual(
            manager._udid_target_map,
            {
                "udid": TargetDescription(
                    udid="udid",
                    name="name",
                    state="state",
                    target_type="type",
                    os_version="os_version",
                    architecture="architecture",
                    companion_info=None,
                    screen_dimensions=None,
                )
            },
        )



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
import sys
from typing import Dict, List, Optional

from idb.grpc.types import CompanionClient
from idb.grpc.stream import Stream
from idb.grpc.idb_pb2 import LaunchRequest, LaunchResponse


Start = LaunchRequest.Start
Stop = LaunchRequest.Stop


async def _drain_stream(stream: Stream[LaunchRequest, LaunchResponse]) -> None:
    async for message in stream:
        pipe = message.pipe
        if pipe:
            if message.interface == LaunchResponse.STDOUT:
                sys.stdout.buffer.write(pipe.data)
                sys.stdout.buffer.flush()
            elif message.interface == LaunchResponse.STDERR:
                sys.stderr.buffer.write(pipe.data)
                sys.stderr.buffer.flush()


async def _end_stream(
    stream: Stream[LaunchRequest, LaunchResponse], stop: asyncio.Event
) -> None:
    await stop.wait()
    await stream.send_message(LaunchRequest(stop=Stop()), end=True)


async def client(
    client: CompanionClient,
    bundle_id: str,
    args: Optional[List[str]] = None,
    env: Optional[Dict[str, str]] = None,
    foreground_if_running: bool = False,
    stop: Optional[asyncio.Event] = None,
) -> None:
    async with client.stub.launch.open() as stream:
        request = LaunchRequest(
            start=Start(
                bundle_id=bundle_id,
                env=env,
                app_args=args,
                foreground_if_running=foreground_if_running,
                wait_for=True if stop else False,
            )
        )
        if stop:
            await stream.send_message(request)
            await asyncio.gather(_drain_stream(stream), _end_stream(stream, stop))
        else:
            await stream.send_message(request, end=True)
            await _drain_stream(stream)



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
from enum import Enum
from io import StringIO
from abc import ABCMeta

from typing import List, NamedTuple, Optional, Set, Union, Dict, Tuple, AsyncIterator

LoggingMetadata = Dict[str, Optional[Union[str, List[str], int, float]]]


class Address(NamedTuple):
    host: str
    grpc_port: int
    port: Optional[int] = None


class AppProcessState(Enum):
    UNKNOWN = 0
    NOT_RUNNING = 1
    RUNNING = 2


class InstalledAppInfo(NamedTuple):
    bundle_id: str
    name: str
    architectures: Set[str]
    install_type: str
    process_state: AppProcessState
    debuggable: bool


class HIDButtonType(Enum):
    APPLE_PAY = 1
    HOME = 2
    LOCK = 3
    SIDE_BUTTON = 4
    SIRI = 5


ConnectionDestination = Union[str, Address]


class CompanionInfo(NamedTuple):
    udid: str
    host: str
    port: int
    is_local: bool
    grpc_port: int


class ScreenDimensions(NamedTuple):
    width: int
    height: int
    density: Optional[float]
    width_points: Optional[int]
    height_points: Optional[int]


class TargetDescription(NamedTuple):
    udid: str
    name: str
    state: Optional[str]
    target_type: Optional[str]
    os_version: Optional[str]
    architecture: Optional[str]
    companion_info: Optional[CompanionInfo]
    screen_dimensions: Optional[ScreenDimensions]


class DaemonInfo(NamedTuple):
    host: str
    port: int
    targets: List[TargetDescription]


ConnectResponse = Union[CompanionInfo, DaemonInfo]


class FileEntryInfo(NamedTuple):
    path: str


class IdbException(Exception):
    pass


class AccessibilityInfo(NamedTuple):
    json: Optional[str]


class CrashLogInfo(NamedTuple):
    name: Optional[str]
    bundle_id: Optional[str]
    process_name: Optional[str]
    parent_process_name: Optional[str]
    process_identifier: Optional[int]
    parent_process_identifier: Optional[int]
    timestamp: Optional[int]


class CrashLog(NamedTuple):
    info: Optional[CrashLogInfo]
    contents: Optional[str]


class CrashLogQuery(NamedTuple):
    since: Optional[int] = None
    before: Optional[int] = None
    bundle_id: Optional[str] = None
    name: Optional[str] = None


class TestRunFailureInfo(NamedTuple):
    message: str
    file: str
    line: int


class TestActivity(NamedTuple):
    title: str
    duration: float
    uuid: str


class TestRunInfo(NamedTuple):
    bundle_name: str
    class_name: str
    method_name: str
    logs: List[str]
    duration: float
    passed: bool
    failure_info: Optional[TestRunFailureInfo]
    activityLogs: Optional[List[TestActivity]]
    crashed: bool


class InstalledTestInfo(NamedTuple):
    bundle_id: str
    name: Optional[str]
    architectures: Optional[Set[str]]


class IdbClientBase:
    async def list_apps(self) -> List[InstalledAppInfo]:
        pass

    async def launch(
        self,
        bundle_id: str,
        env: Optional[Dict[str, str]] = None,
        args: Optional[List[str]] = None,
        foreground_if_running: bool = False,
        stop: Optional[asyncio.Event] = None,
    ) -> None:
        pass

    async def run_xctest(
        self,
        test_bundle_id: str,
        app_bundle_id: str,
        test_host_app_bundle_id: Optional[str] = None,
        is_ui_test: bool = False,
        is_logic_test: bool = False,
        tests_to_run: Optional[Set[str]] = None,
        tests_to_skip: Optional[Set[str]] = None,
        env: Optional[Dict[str, str]] = None,
        args: Optional[List[str]] = None,
        result_bundle_path: Optional[str] = None,
        idb_log_buffer: Optional[StringIO] = None,
        timeout: Optional[int] = None,
        poll_interval_sec: float = 0.5,
    ) -> AsyncIterator[TestRunInfo]:
        yield

    async def install(self, bundle_path: str) -> str:
        pass

    async def uninstall(self, bundle_id: str) -> None:
        pass

    async def install_dylib(self, dylib_path: str) -> str:
        pass

    async def install_xctest(self, bundle_path: str) -> str:
        pass

    async def connect(
        self,
        destination: Union[Address, str],
        metadata: Optional[Dict[str, str]] = None,
    ) -> None:
        return

    async def disconnect(self, destination: Union[Address, str]) -> None:
        pass

    async def list_targets(self) -> List[TargetDescription]:
        pass

    async def list_xctests(self) -> List[InstalledTestInfo]:
        pass

    async def terminate(self, bundle_id: str) -> None:
        pass

    async def pull(self, bundle_id: str, src_path: str, dest_path: str) -> None:
        pass

    async def mkdir(self, bundle_id: str, path: str) -> None:
        pass

    async def list_test_bundle(self, test_bundle_id: str) -> List[str]:
        pass

    async def tail_logs(
        self, stop: asyncio.Event, arguments: Optional[List[str]] = None
    ) -> AsyncIterator[str]:
        yield

    async def push(self, src_paths: List[str], bundle_id: str, dest_path: str) -> None:
        pass

    async def clear_keychain(self) -> None:
        pass

    async def boot(self) -> None:
        pass

    async def open_url(self, url: str) -> None:
        pass

    async def set_location(self, latitude: float, longitude: float) -> None:
        pass

    async def approve(self, bundle_id: str, permissions: Set[str]) -> None:
        pass

    async def record_video(self, stop: asyncio.Event, output_file: str) -> None:
        pass

    async def screenshot(self) -> bytes:
        pass

    async def tap(self, x: int, y: int, duration: Optional[float] = None) -> None:
        pass

    async def button(
        self, button_type: HIDButtonType, duration: Optional[float] = None
    ) -> None:
        pass

    async def key(self, keycode: int, duration: Optional[float] = None) -> None:
        return

    async def key_sequence(self, key_sequence: List[int]) -> None:
        pass

    async def swipe(
        self,
        p_start: Tuple[int, int],
        p_end: Tuple[int, int],
        delta: Optional[int] = None,
    ) -> None:
        pass

    async def crash_show(self, name: str) -> CrashLog:
        pass

    async def contacts_update(self, contacts_path: str) -> None:
        pass

    async def describe(self) -> TargetDescription:
        pass

    async def accessibility_info(
        self, point: Optional[Tuple[int, int]]
    ) -> AccessibilityInfo:
        pass

    async def run_instruments(
        self,
        stop: asyncio.Event,
        template: str,
        app_bundle_id: str,
        trace_path: str,
        post_process_arguments: Optional[List[str]] = None,
        env: Optional[Dict[str, str]] = None,
        app_args: Optional[List[str]] = None,
        started: Optional[asyncio.Event] = None,
    ) -> None:
        return

    async def start_instruments(
        self,
        template: str,
        app_bundle_id: str,
        env: Optional[Dict[str, str]] = None,
        app_args: Optional[List[str]] = None,
    ) -> Optional[str]:
        return

    async def stop_instruments(
        self, session_id: str, post_process_arguments: Optional[List[str]] = None
    ) -> Tuple[Optional[bytes], Optional[str]]:
        pass

    async def crash_list(self, query: CrashLogQuery) -> List[CrashLogInfo]:
        pass

    async def crash_delete(self, query: CrashLogQuery) -> List[CrashLogInfo]:
        pass

    async def add_metadata(self, metadata: Optional[Dict[str, str]]) -> None:
        pass

    async def add_media(self, file_paths: List[str]) -> None:
        pass

    async def focus(self) -> None:
        pass

    async def tail_logs_contextmanager(self) -> AsyncIterator[str]:
        yield

    async def debugserver_start(self, bundle_id: str) -> List[str]:
        pass

    async def debugserver_stop(self) -> None:
        pass

    async def debugserver_status(self) -> Optional[List[str]]:
        pass

    async def text(self, text: str) -> None:
        return

    async def ls(self, bundle_id: str, path: str) -> List[FileEntryInfo]:
        pass

    async def mv(self, bundle_id: str, src_paths: List[str], dest_path: str) -> None:
        pass

    async def rm(self, bundle_id: str, paths: List[str]) -> None:
        pass


class Server(metaclass=ABCMeta):
    def close(self) -> None:
        pass

    async def wait_closed(self) -> None:
        pass

    @property
    def ports(self) -> Dict[str, str]:
        pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import os
import plistlib
from enum import Enum
from typing import Any, Dict, List


class XCTestException(Exception):
    pass


class _XCTestType(Enum):
    XCTest = 0
    XCTestRun = 1


def _get_xctest_type(path: str) -> _XCTestType:
    if path.endswith(".xctestrun") and os.path.isfile(path):
        return _XCTestType.XCTestRun
    if path.endswith(".xctest") and os.path.isdir(path):
        return _XCTestType.XCTest
    raise XCTestException(f"{path} is not a valid xctest target")


def extract_paths_from_xctestrun(path: str) -> List[str]:
    """
    When using xctestrun we need to copy:
    - the xctestrun file
    - the host app directory that is specified in the xctestrun file for
    every test.
    This method returns paths to those.
    """
    result = [path]
    test_root = os.path.dirname(path)
    with open(path, "rb") as f:
        xctestrun_dict: Dict[str, Any] = plistlib.load(f, use_builtin_types=True)
        for _test_id, test_dict in xctestrun_dict.items():
            result.append(test_dict["TestHostPath"].replace("__TESTROOT__", test_root))
    return result


def xctest_paths_to_tar(bundle_path: str) -> List[str]:
    test_type = _get_xctest_type(bundle_path)
    if test_type is _XCTestType.XCTest:
        return [bundle_path]
    with open(bundle_path, "rb") as f:
        plist: Dict[str, Any] = plistlib.load(f)
        use_artifacts = (
            v.get("UseDestinationArtifacts", False) is True for v in plist.values()
        )

    if not all(use_artifacts):
        return extract_paths_from_xctestrun(bundle_path)
    else:
        return [bundle_path]



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

import asyncio
import importlib.util
import inspect
import pkgutil
from functools import partial, wraps
from inspect import signature
from types import ModuleType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
)

from grpclib.exceptions import GRPCError, ProtocolError, StreamTerminatedError
from idb.common.boot_manager import BootManager
from idb.grpc.types import CompanionClient
from idb.manager.companion import CompanionManager
from idb.grpc.stream import Stream
from idb.common.logging import log_call
from idb.common.types import IdbException, LoggingMetadata
from idb.grpc.idb_grpc import CompanionServiceStub


"""
This module is responsible for getting client and daemon calls from a set of modules
then providing them as functions that can be annotated into client & daemon classes.

It does this by introspecting all modules within the idb.ipc module
and then traversing them to find the relevant implementations.

This is all vended to consumers as a Tuple[call_name, implementation] that can
then be installed using setattr.

For client calls:
- The name of the module is used as the name of the method.
- The function in the module is called 'client'.
- Modules can also provide extra properties to be attached to the client, for
  the purpose of command aliasing. This can be done by setting CLIENT_PROPERTIES
  to a list of functions

For daemon calls:
- The name of the module is used as the name of the method.
- The function in the module is called 'daemon'
"""

BASE_PACKAGE = "idb.ipc"


class DaemonContext(NamedTuple):
    companion_manager: CompanionManager
    boot_manager: BootManager


CompanionProvider = Callable[[Optional[str]], Awaitable[CompanionClient]]
DaemonContextProvider = Callable[[], Awaitable[DaemonContext]]
DaemonProvider = Callable[[], Awaitable[CompanionClient]]
_T = TypeVar("_T")
_U = TypeVar("_U")


CLIENT_METADATA: LoggingMetadata = {"component": "client", "rpc_protocol": "grpc"}
DAEMON_METADATA: LoggingMetadata = {"component": "daemon", "rpc_protocol": "grpc"}


class MetadataStubInjector(CompanionServiceStub):
    def __init__(self, stub: CompanionServiceStub, metadata: Dict[str, str]) -> None:
        self._stub = stub
        self._metadata = metadata

    def __getattr__(self, name: str) -> Any:  # pyre-ignore
        if not hasattr(self._stub, name):
            return getattr(super(), name)

        call = getattr(self._stub, name)
        metadata = self._metadata

        class StubTrampoline:
            def open(self, *args: Any, **kwargs: Any) -> Any:  # pyre-ignore
                return call.open(*args, **kwargs, metadata=metadata)

            def __call__(self, *args: Any, **kwargs: Any) -> Any:  # pyre-ignore
                return call(*args, **kwargs, metadata=metadata)

        return StubTrampoline()


def _trampoline_client(
    daemon_provider: DaemonProvider, call: Callable, name: str
) -> Callable:
    async def _make_client() -> CompanionClient:
        client = await daemon_provider()
        return CompanionClient(
            stub=MetadataStubInjector(
                stub=client.stub, metadata={"udid": client.udid} if client.udid else {}
            ),
            is_local=client.is_local,
            udid=client.udid,
            logger=client.logger.getChild(name),
        )

    @log_call(name=name, metadata=CLIENT_METADATA)
    @wraps(call)
    async def _tramp(*args: Any, **kwargs: Any) -> Any:  # pyre-ignore
        try:
            client = await _make_client()
            return await call(client, *args, **kwargs)
        except GRPCError as e:
            raise IdbException(e.message) from e  # noqa B306
        except (ProtocolError, StreamTerminatedError) as e:
            raise IdbException(e.args) from e

    @log_call(name=name, metadata=CLIENT_METADATA)
    @wraps(call)
    async def _tramp_gen(*args: Any, **kwargs: Any) -> Any:  # pyre-ignore
        try:
            client = await _make_client()
            async for item in call(client, *args, **kwargs):
                yield item
        except GRPCError as e:
            raise IdbException(e.message) from e  # noqa B306
        except (ProtocolError, StreamTerminatedError) as e:
            raise IdbException(e.args) from e

    if inspect.isasyncgenfunction(call):
        return _tramp_gen
    else:
        return _tramp


def _default_daemon(
    name: str,
) -> Callable[[CompanionClient, Stream[_T, _U]], Awaitable[None]]:
    async def _pipe_to_companion(
        in_stream: Stream[_T, _U],
        out_stream: Stream[_U, _T],
        started_future: asyncio.Future,
    ) -> None:
        async for message in in_stream:
            await out_stream.send_message(message)
            if not started_future.done():
                started_future.set_result(None)
        await out_stream.end()

    async def _pipe_to_client(
        in_stream: Stream[_U, _T],
        out_stream: Stream[_T, _U],
        started_future: asyncio.Future,
    ) -> None:
        await started_future
        async for message in in_stream:
            await out_stream.send_message(message)

    async def _default_daemon_imp(
        client: CompanionClient, stream: Stream[_T, _U]
    ) -> None:
        method = getattr(client.stub, name)
        async with method.open() as out_stream:
            started_future = asyncio.Future()
            await asyncio.gather(
                _pipe_to_companion(stream, out_stream, started_future),
                _pipe_to_client(out_stream, stream, started_future),
            )

    return _default_daemon_imp


def _trampoline_daemon(
    companion_provider: CompanionProvider,
    context_provider: DaemonContextProvider,
    call: Callable,
    name: str,
) -> Callable:
    @log_call(name=name, metadata=DAEMON_METADATA, translate_exceptions=True)
    @wraps(call)
    async def _tramp(stream: Stream[_T, _U], *args: Any, **kwargs: Any) -> None:
        partial_call = call
        if _takes_client(call):
            client = await companion_provider(stream.metadata.get("udid"))
            logger = client.logger.getChild(name)
            companion_client = CompanionClient(
                stub=client.stub,
                is_local=client.is_local,
                udid=client.udid,
                logger=logger,
                is_companion_available=client.is_companion_available,
            )
            partial_call = partial(partial_call, client=companion_client)
        if _takes_context(call):
            context = await context_provider()
            partial_call = partial(partial_call, context=context)

        if _takes_stream(call):
            await partial_call(stream=stream)
        else:
            request = await stream.recv_message()
            response = await partial_call(request=request)
            await stream.send_message(response)

    return _tramp


def _takes_stream(method: Callable) -> bool:
    return _has_parameter(method=method, name="stream", parameter_type=Stream)


def _takes_client(method: Callable) -> bool:
    return _has_parameter(method=method, name="client", parameter_type=CompanionClient)


def _takes_context(method: Callable) -> bool:
    return _has_parameter(method=method, name="context", parameter_type=DaemonContext)


def _has_parameter(
    method: Callable, name: str, parameter_type: Optional[Type[_T]] = None
) -> bool:
    parameters = list(signature(method).parameters.items())
    return any(
        (
            parameter_name == name
            and (
                parameter_type is None
                or is_subclass(parameter.annotation, parameter_type)
            )
            for parameter_name, parameter in parameters
        )
    )


def is_subclass(obj: Type[_T], target_type: Type[_U]) -> bool:
    obj = getattr(obj, "__origin__", obj) or obj
    target_type = getattr(target_type, "__origin__", target_type) or target_type
    return issubclass(obj, target_type)


def _get_rpc_modules(base_package: str = BASE_PACKAGE) -> List[ModuleType]:
    package = importlib.util.find_spec(base_package)
    if package is None:
        return []
    return [
        importlib.import_module(f"{base_package}.{name}")
        for (_, name, ispkg) in pkgutil.iter_modules(package.submodule_search_locations)
        if not ispkg
    ]


def _get_module_name(module: ModuleType) -> str:
    return module.__name__.split(".")[-1]


def client_calls(daemon_provider: DaemonProvider) -> List[Tuple[str, Callable]]:
    return [
        (
            name,
            _trampoline_client(daemon_provider=daemon_provider, call=call, name=name),
        )
        for (name, call) in _client_implementations()
    ]


def _client_implementations() -> List[Tuple[str, Callable]]:
    properties = []
    for module in _get_rpc_modules():
        if hasattr(module, "client"):
            properties.append((_get_module_name(module), module.client))  # pyre-ignore
        for extra_property in getattr(module, "CLIENT_PROPERTIES", []):
            properties.append((extra_property.__name__, extra_property))
    return properties


def daemon_calls(
    companion_provider: CompanionProvider, context_provider: DaemonContextProvider
) -> List[Tuple[str, Callable]]:
    return [
        (
            name,
            _trampoline_daemon(
                companion_provider=companion_provider,
                context_provider=context_provider,
                call=call,
                name=name,
            ),
        )
        for (name, call) in _daemon_implementations()
    ]


def _daemon_implementations() -> List[Tuple[str, Callable]]:
    functions = []
    for module in _get_rpc_modules():
        module_name = _get_module_name(module)
        call = getattr(module, "daemon", _default_daemon(module_name))
        functions.append((module_name, call))
    return functions
